--******************************************************************************
--* Company:     FTDI                                                          *
--* Date:        12 Feb 2004                                                   *
--* Author:      Iain Maciver & John Quinn                                     *
--*                                                                            *
--* Module Name: mcontrol                                                      *
--*                                                                            *
--* Description: Main Control block for morphic I/O control.                   *
--*              Controls transfer of data between FT232 FIFO and this module. *
--*              Commands interpreted are Read, Write, change Output Enables   *
--*              and Set Clk.                                                  *
--*                                                                            *
--* Sub-modules: none                                                          *
--******************************************************************************

LIBRARY IEEE;
USE IEEE.std_logic_1164.ALL;

ENTITY mcontrol_mII IS
   PORT(
        -- inputs
        clk     : in std_logic;
        rst     : in std_logic;
        clkin   : in std_logic;
        cnt     : in std_logic_vector(11 downto 0);
        nRXF    : in std_logic;
        nTXE    : in std_logic;
        CMDData : in std_logic_vector(7 downto 0);
        IOAin   : in std_logic_vector(7 downto 0);
        IOBin   : in std_logic_vector(7 downto 0);
        IOCin   : in std_logic_vector(7 downto 0);
        IODin   : in std_logic_vector(7 downto 0);
        IOEin   : in std_logic_vector(7 downto 0);
        IOFin   : in std_logic_vector(7 downto 0);
        IOGin   : in std_logic_vector(7 downto 0);
        IOHin   : in std_logic_vector(7 downto 0);
        IOIin   : in std_logic_vector(7 downto 0);
        IOJin   : in std_logic_vector(7 downto 0);

        -- outputs
        FT232Dout : out std_logic_vector(7 downto 0);
        FT232DEn  : out std_logic;
        nRD       : out std_logic;
        nWR       : out std_logic;
        oe_n      : out std_logic;
        IOAout    : out std_logic_vector(7 downto 0);
        IOBout    : out std_logic_vector(7 downto 0);
        IOCout    : out std_logic_vector(7 downto 0);
        IODout    : out std_logic_vector(7 downto 0);
        IOEout    : out std_logic_vector(7 downto 0);
        IOFout    : out std_logic_vector(7 downto 0);
        IOGout    : out std_logic_vector(7 downto 0);
        IOHout    : out std_logic_vector(7 downto 0);
        IOIout    : out std_logic_vector(7 downto 0);
        IOJout    : out std_logic_vector(7 downto 0);
        IOAen     : out std_logic_vector(7 downto 0);
        IOBen     : out std_logic_vector(7 downto 0);
        IOCen     : out std_logic_vector(7 downto 0);
        IODen     : out std_logic_vector(7 downto 0);
        IOEen     : out std_logic_vector(7 downto 0);
        IOFen     : out std_logic_vector(7 downto 0);
        IOGen     : out std_logic_vector(7 downto 0);
        IOHen     : out std_logic_vector(7 downto 0);
        IOIen     : out std_logic_vector(7 downto 0);
        IOJen     : out std_logic_vector(7 downto 0)
        );
END mcontrol_mII;

ARCHITECTURE rtl OF mcontrol_mII IS

   attribute enum_encoding : string;

   type rxstat is (RxIDLE, RxCMD1, RxCMD2, RxCMD3, RxDATA1, RxDATA2);
   type txstat is (TxIDLE, TxWAIT1, TxCMD1, TxCMD2, TxWAIT2, TxDATA1, TxDATA2);

   -- used if user-encoded scheme is chosen in synthesiser for state encoding
   attribute enum_encoding of rxstat : type is "000 001 010 100 101 110";
   attribute enum_encoding of txstat : type is "000 001 010 100 101 110 111";

   signal RxState     : rxstat;          -- Curent state of Rx State Machine
   signal TxState     : txstat;          -- Curent state of Tx State Machine

   signal intRXF      : std_logic;       -- Rx FIFO full
   signal intTXE      : std_logic;       -- Tx FIFO empty
   signal intRD       : std_logic;       -- Read data from FIFO
   signal LdCmdReg    : std_logic;       -- Store CMDData command in CmdReg
   signal LdDataReg   : std_logic;       -- Store CMDData data in DataReg
   signal LoadStb     : std_logic;       -- Update an output or enable reg
--    signal LdOutReg    : std_logic;       -- Update an output register
--    signal LdEnReg     : std_logic;       -- Update an enable register
   signal CmdReg      : std_logic_vector(7 downto 0); -- stores prev command /
   signal DataReg     : std_logic_vector(7 downto 0); -- stores prev data
   signal senddata    : std_logic;       -- Send port data to FT232Dout
   signal loadFT232D  : std_logic;       -- Update output data for a read
   signal loadIOAout  : std_logic;       -- Update IOAout reg
   signal loadIOBout  : std_logic;       --   "    IOBout  "
   signal loadIOCout  : std_logic;       --   "    IOCout  "
   signal loadIODout  : std_logic;       --   "    IODout  "
   signal loadIOEout  : std_logic;       --   "    IOEout  "
   signal loadIOFout  : std_logic;       --   "    IOFout  "
   signal loadIOGout  : std_logic;       --   "    IOGout  "
   signal loadIOHout  : std_logic;       --   "    IOHout  "
   signal loadIOIout  : std_logic;       --   "    IOIout  "
   signal loadIOJout  : std_logic;       --   "    IOJout  "
   signal loadIOAen   : std_logic;       -- Update IOAen reg
   signal loadIOBen   : std_logic;       --   "    IOBen  "
   signal loadIOCen   : std_logic;       --   "    IOCen  "
   signal loadIODen   : std_logic;       --   "    IODen  "
   signal loadIOEen   : std_logic;       --   "    IOEen  "
   signal loadIOFen   : std_logic;       --   "    IOFen  "
   signal loadIOGen   : std_logic;       --   "    IOGen  "
   signal loadIOHen   : std_logic;       --   "    IOHen  "
   signal loadIOIen   : std_logic;       --   "    IOIen  "
   signal loadIOJen   : std_logic;       --   "    IOJen  "
   -- Internal Input, Output and Output Enable signals
   signal intIOAin    : std_logic_vector(7 downto 0);
   signal intIOBin    : std_logic_vector(7 downto 0);
   signal intIOCin    : std_logic_vector(7 downto 0);
   signal intIODin    : std_logic_vector(7 downto 0);
   signal intIOEin    : std_logic_vector(7 downto 0);
   signal intIOFin    : std_logic_vector(7 downto 0);
   signal intIOGin    : std_logic_vector(7 downto 0);
   signal intIOHin    : std_logic_vector(7 downto 0);
   signal intIOIin    : std_logic_vector(7 downto 0);
   signal intIOJin    : std_logic_vector(7 downto 0);
   signal intIOAout   : std_logic_vector(7 downto 0);
   signal intIOBout   : std_logic_vector(7 downto 0);
   signal intIOCout   : std_logic_vector(7 downto 0);
   signal intIODout   : std_logic_vector(7 downto 0);
   signal intIOEout   : std_logic_vector(7 downto 0);
   signal intIOFout   : std_logic_vector(7 downto 0);
   signal intIOGout   : std_logic_vector(7 downto 0);
   signal intIOHout   : std_logic_vector(7 downto 0);
   signal intIOIout   : std_logic_vector(7 downto 0);
   signal intIOJout   : std_logic_vector(7 downto 0);
   signal intIOAen    : std_logic_vector(7 downto 0);
   signal intIOBen    : std_logic_vector(7 downto 0);
   signal intIOCen    : std_logic_vector(7 downto 0);
   signal intIODen    : std_logic_vector(7 downto 0);
   signal intIOEen    : std_logic_vector(7 downto 0);
   signal intIOFen    : std_logic_vector(7 downto 0);
   signal intIOGen    : std_logic_vector(7 downto 0);
   signal intIOHen    : std_logic_vector(7 downto 0);
   signal intIOIen    : std_logic_vector(7 downto 0);
   signal intIOJen    : std_logic_vector(7 downto 0);

   signal intFT232DEn  : std_logic;
   signal intFT232Dout : std_logic_vector(7 downto 0);
   signal int_Din      : std_logic_vector(7 downto 0);
   signal port_sel     : std_logic_vector(9 downto 0); -- Ctl for port to o/p
   signal clk_selA     : std_logic_vector(3 downto 0); -- IOA Clk o/p control
   signal clk_selB     : std_logic_vector(3 downto 0); -- IOB Clk o/p control
   signal clk_selC     : std_logic_vector(3 downto 0); -- IOC Clk o/p control
   signal clk_selD     : std_logic_vector(3 downto 0); -- IOD Clk o/p control
   signal clk_selE     : std_logic_vector(3 downto 0); -- IOE Clk o/p control
   signal clk_selF     : std_logic_vector(3 downto 0); -- IOF Clk o/p control
   signal clk_selG     : std_logic_vector(3 downto 0); -- IOG Clk o/p control
   signal clk_selH     : std_logic_vector(3 downto 0); -- IOH Clk o/p control
   signal clk_selI     : std_logic_vector(3 downto 0); -- IOI Clk o/p control
   signal clk_selJ     : std_logic_vector(3 downto 0); -- IOJ Clk o/p control

   signal DecodeRead  : std_logic;       -- CmdReg decode indicates Read
   signal DecodeWrite : std_logic;       --   "      "        "     Write
   signal DecodeOEn   : std_logic;       --   "      "        "     OEn
   signal DecodeClk   : std_logic;       --   "      "        "     Clk
   signal DecodeIOA   : std_logic;       --   "      "        "     IOA
   signal DecodeIOB   : std_logic;       --   "      "        "     IOB
   signal DecodeIOC   : std_logic;       --   "      "        "     IOC
   signal DecodeIOD   : std_logic;       --   "      "        "     IOD
   signal DecodeIOE   : std_logic;       --   "      "        "     IOE
   signal DecodeIOF   : std_logic;       --   "      "        "     IOF
   signal DecodeIOG   : std_logic;       --   "      "        "     IOG
   signal DecodeIOH   : std_logic;       --   "      "        "     IOH
   signal DecodeIOI   : std_logic;       --   "      "        "     IOI
   signal DecodeIOJ   : std_logic;       --   "      "        "     IOJ
   signal LoadCommand : std_logic;       -- Decoded command is valid
--    signal DecodeSync  : std_logic;       --   "      "      not recognised
   signal SyncOver    : std_logic;       -- CmdReg decode indicates sync is over
   signal SyncStart   : std_logic;       --   "      "        "    start of sync
   signal sync_mode   : std_logic;       -- Indicates when in Sync Mode

BEGIN

--------------------------------------------------------------------------------
-- Register the inputs
--------------------------------------------------------------------------------
   regip : process (clk, rst)
   begin
   if (rst = '1') then
      intIOAin <= "00000000";
      intIOBin <= "00000000";
      intIOCin <= "00000000";
      intIODin <= "00000000";
      intIOEin <= "00000000";
      intIOFin <= "00000000";
      intIOGin <= "00000000";
      intIOHin <= "00000000";
      intIOIin <= "00000000";
      intIOJin <= "00000000";
      intRXF   <= '0';
      intTXE   <= '0';
   elsif ((clk'EVENT) and (clk='1')) then
      intIOAin <= IOAin;
      intIOBin <= IOBin;
      intIOCin <= IOCin;
      intIODin <= IODin;
      intIOEin <= IOEin;
      intIOFin <= IOFin;
      intIOGin <= IOGin;
      intIOHin <= IOHin;
      intIOIin <= IOIin;
      intIOJin <= IOJin;
      intRXF   <= not nRXF;
      intTXE   <= not nTXE;
   end if;
   end process regip;

--------------------------------------------------------------------------------
-- Hold Command and Data Instructions
--------------------------------------------------------------------------------
   regcd : process (clk, rst)
   begin
   if (rst = '1') then
      CmdReg  <= "00000000";
      DataReg <= "00000000";
   elsif ((clk'EVENT) and (clk='1')) then
      -- store input data in CmdReg when LdCmdReg is high
      if (LdCmdReg = '1') then
          CmdReg <= CMDData;
      else
          CmdReg <= CmdReg;
      end if;

      -- store input data in DataReg when LdDataReg is high
      if (LdDataReg = '1') then
          DataReg <= CMDData;
      else
          DataReg <= DataReg;
      end if;
   end if;
   end process regcd;

--------------------------------------------------------------------------------
-- Decode the Commands: Read  -> CmdReg(7)
--                      Write -> CmdReg(6)
--                      Oen   -> CmdReg(5)
--                      Clk   -> CmdReg(6) and CmdReg(5)
--                      --
--                      IOA   -> CmdReg(0)
--                      IOB   -> CmdReg(1)
--                      IOC   -> CmdReg(1) and CmdReg(0)
--                      IOD   -> CmdReg(2)
--                      IOE   -> CmdReg(2) and CmdReg(0)
--                      IOF   -> CmdReg(2) and CmdReg(1)
--                      IOG   -> CmdReg(2) and CmdReg(1) and CmdReg(0)
--                      IOH   -> CmdReg(3)
--                      IOI   -> CmdReg(3) and CmdReg(0)
--                      IOJ   -> CmdReg(3) and CmdReg(1)
--                      --
--                      Sync  -> Command not recognised therefore enter sync
--------------------------------------------------------------------------------
   -- DecodeRead <= '1' when CmdReg(7 downto 5) = "100" else '0';

   DecodeRead  <=     CmdReg(7) and
                  not CmdReg(6) and
                  not CmdReg(5);

   DecodeWrite <= not CmdReg(7) and
                      CmdReg(6) and
                  not CmdReg(5);

   DecodeOEn   <= not CmdReg(7) and
                  not CmdReg(6) and
                      CmdReg(5);

   DecodeClk   <= not CmdReg(7) and
                      CmdReg(6) and
                      CmdReg(5);

   DecodeIOA   <= not CmdReg(4) and
                 not CmdReg(3) and
                 not CmdReg(2) and
                 not CmdReg(1) and
                      CmdReg(0);

   DecodeIOB   <= not CmdReg(4) and
                  not CmdReg(3) and
                  not CmdReg(2) and
                      CmdReg(1) and
                  not CmdReg(0);

   DecodeIOC   <= not CmdReg(4) and
                  not CmdReg(3) and
                  not CmdReg(2) and
                      CmdReg(1) and
                      CmdReg(0);

   DecodeIOD   <= not CmdReg(4) and
                  not CmdReg(3) and
                      CmdReg(2) and
                  not CmdReg(1) and
                  not CmdReg(0);

   DecodeIOE   <= not CmdReg(4) and
                  not CmdReg(3) and
                      CmdReg(2) and
                  not CmdReg(1) and
                      CmdReg(0);

   DecodeIOF   <= not CmdReg(4) and
                  not CmdReg(3) and
                      CmdReg(2) and
                      CmdReg(1) and
                  not CmdReg(0);

   DecodeIOG   <= not CmdReg(4) and
                  not CmdReg(3) and
                      CmdReg(2) and
                      CmdReg(1) and
                      CmdReg(0);

   DecodeIOH   <= not CmdReg(4) and
                      CmdReg(3) and
                  not CmdReg(2) and
                  not CmdReg(1) and
                  not CmdReg(0);

   DecodeIOI   <= not CmdReg(4) and
                      CmdReg(3) and
                  not CmdReg(2) and
                  not CmdReg(1) and
                      CmdReg(0);

   DecodeIOJ   <= not CmdReg(4) and
                      CmdReg(3) and
                  not CmdReg(2) and
                      CmdReg(1) and
                  not CmdReg(0);


   -- Sync Mode begins when a command is not recognised
   SyncStart   <= not LoadCommand and not sync_mode;

   -- Sync Mode is over when a Command of 0x140B is received
   SyncOver    <=     CmdReg(7) and
                  not CmdReg(6) and
                      CmdReg(5) and
                  not CmdReg(4) and
                      CmdReg(3) and
                  not CmdReg(2) and
                      CmdReg(1) and
                      CmdReg(0);

   -- Used to set LoadStb if the command is valid and recognised
   LoadCommand <= (DecodeRead or DecodeWrite or DecodeOEn or DecodeClk) and
                  (DecodeIOA or DecodeIOB or DecodeIOC or DecodeIOD or
                   DecodeIOE or DecodeIOF or DecodeIOG or DecodeIOH or DecodeIOI or DecodeIOJ)
                   and not sync_mode;

--------------------------------------------------------------------------------
-- Rx State Machine control
--------------------------------------------------------------------------------
   RxStatep : process (clk, rst)
   begin
   if (rst = '1') then
      RxState   <= RxIDLE;
      intRD     <= '0';
      LdCmdReg  <= '0';
      LdDataReg <= '0';
      LoadStb   <= '0';
      sync_mode <= '0';
   elsif ((clk'EVENT) and (clk='1')) then
      case RxState is
         -- Wait for Rx FIFO to signal that it has data to send
         -- (this should be Command data - LdCmdReg set high)
         when RxIDLE =>
            if (intRXF = '1') then
               RxState  <= RxCMD1;
               intRD    <= '1';
               LdCmdReg <= '1';
            else
               RxState  <= RxIDLE;
               intRD    <= '0';
               LdCmdReg <= '0';
            end if;
            LdDataReg <= '0';
            LoadStb   <= '0';
            sync_mode <= sync_mode;
         -- RD is set high for two cycles to read the data from FIFO
         when RxCMD1 =>
             RxState   <= RxCMD2;
             intRD     <= '1';
             LdCmdReg  <= '1';
             LdDataReg <= '0';
             LoadStb   <= '0';
             sync_mode <= sync_mode;
         -- Remain in this state until FIFO indicates Rd cycle is complete
         -- When this occurs test to see whether synchroisation needs to occur
         -- and if it does remain in sync mode until Cmd of 0x140B is received
         when RxCMD2 =>
            if (intRXF = '1') then
               RxState   <= RxCMD2;
               LoadStb   <= '0';
               sync_mode <= sync_mode;
            elsif (LoadCommand = '0') then
               RxState   <= RxIDLE;
               LoadStb   <= '1';
               sync_mode <= SyncStart or not SyncOver;
            else
               RxState   <= RxCMD3;
               LoadStb   <= '0';
               sync_mode <= sync_mode;
            end if;
            intRD     <= '0';
            LdCmdReg  <= '0';
            LdDataReg <= '0';
         -- Again wait until Rx FIFO has data to send
         -- (this time it should be a Data word - LdDataReg set high)
         when RxCMD3 =>
            if (intRXF = '1') then
               RxState   <= RxDATA1;
               intRD     <= '1';
               LdDataReg <= '1';
            else
               RxState   <= RxCMD3;
               intRD     <= '0';
               LdDataReg <= '0';
            end if;
            LdCmdReg  <= '0';
            LoadStb   <= '0';
            sync_mode <= sync_mode;
         -- RD is set high for two cycles to read the data from FIFO
         when RxDATA1 =>
            RxState   <= RxDATA2;
            intRD     <= '1';
            LdCmdReg  <= '0';
            LdDataReg <= '1';
            LoadStb   <= '0';
            sync_mode <= sync_mode;
         -- Remain in this state until FIFO indicates Rd cycle is complete.
         -- When it is, LoadStb is pulsed high if the command and data are
         -- both valid.
         when RxDATA2 =>
            if (intRXF = '1') then
               RxState <= RxDATA2;
               LoadStb <= '0';
            else
               RxState <= RxIDLE;
               if (LoadCommand = '1') then
                  LoadStb <= '1';
               else
                  LoadStb <= '0';
               end if;
            end if;
            intRD     <= '0';
            LdCmdReg  <= '0';
            LdDataReg <= '0';
            sync_mode <= sync_mode;
         -- default state - return to IDLE
         when others =>
            RxState   <= RxIDLE;
            intRD     <= '0';
            LdCmdReg  <= '0';
            LdDataReg <= '0';
            LoadStb   <= '0';
            sync_mode <= sync_mode;
      end case;
   end if;
   end process RxStatep;

--------------------------------------------------------------------------------
-- Clock select registers. Chooses what clk frequency (if any) to output on
-- the Clk outputs. The clock frequency is decoded from 3 downto 0 of Data Reg.
--------------------------------------------------------------------------------
   regclk_sel : process (clk, rst)
   begin
   if (rst = '1') then
      clk_selA <= "0000";
      clk_selB <= "0000";
      clk_selC <= "0000";
      clk_selD <= "0000";
      clk_selE <= "0000";
      clk_selF <= "0000";
      clk_selG <= "0000";
      clk_selH <= "0000";
      clk_selI <= "0000";
      clk_selJ <= "0000";
   elsif ((clk'EVENT) and (clk='1')) then
      if (LoadStb = '1' and DecodeClk = '1') then
         --
         if (DecodeIOA = '1') then
            clk_selA <= DataReg(3 downto 0);
         else
            clk_selA <= clk_selA;
         end if;
         --
         if (DecodeIOB = '1') then
            clk_selB <= DataReg(3 downto 0);
         else
            clk_selB <= clk_selB;
         end if;
         --
         if (DecodeIOC = '1') then
            clk_selC <= DataReg(3 downto 0);
         else
            clk_selC <= clk_selC;
         end if;
         --
         if (DecodeIOD = '1') then
            clk_selD <= DataReg(3 downto 0);
         else
            clk_selD <= clk_selD;
         end if;
         --
         if (DecodeIOE = '1') then
            clk_selE <= DataReg(3 downto 0);
         else
            clk_selE <= clk_selE;
         end if;
         --
          if (DecodeIOF = '1') then
            clk_selF <= DataReg(3 downto 0);
         else
            clk_selF <= clk_selF;
         end if;
         --
          if (DecodeIOG = '1') then
            clk_selG <= DataReg(3 downto 0);
         else
            clk_selG <= clk_selG;
         end if;
         --
          if (DecodeIOH = '1') then
            clk_selH <= DataReg(3 downto 0);
         else
            clk_selH <= clk_selH;
         end if;
         --
          if (DecodeIOI = '1') then
            clk_selI <= DataReg(3 downto 0);
         else
            clk_selI <= clk_selI;
         end if;
         --
          if (DecodeIOJ = '1') then
            clk_selJ <= DataReg(3 downto 0);
         else
            clk_selJ <= clk_selJ;
         end if;
         --
      else
         clk_selA <= clk_selA;
         clk_selB <= clk_selB;
         clk_selC <= clk_selC;
         clk_selD <= clk_selD;
         clk_selE <= clk_selE;
         clk_selF <= clk_selF;
         clk_selG <= clk_selG;
         clk_selH <= clk_selH;
         clk_selI <= clk_selI;
         clk_selJ <= clk_selJ;
      end if;
   end if;
   end process regclk_sel;

--------------------------------------------------------------------------------
-- Registers for Output data and Output Enable signals
--------------------------------------------------------------------------------
   -- Generate control signals for loading data to output registers
   loadIOAout <= LoadStb and DecodeWrite and DecodeIOA;
   loadIOBout <= LoadStb and DecodeWrite and DecodeIOB;
   loadIOCout <= LoadStb and DecodeWrite and DecodeIOC;
   loadIODout <= LoadStb and DecodeWrite and DecodeIOD;
   loadIOEout <= LoadStb and DecodeWrite and DecodeIOE;
   loadIOFout <= LoadStb and DecodeWrite and DecodeIOF;
   loadIOGout <= LoadStb and DecodeWrite and DecodeIOG;
   loadIOHout <= LoadStb and DecodeWrite and DecodeIOH;
   loadIOIout <= LoadStb and DecodeWrite and DecodeIOI;
   loadIOJout <= LoadStb and DecodeWrite and DecodeIOJ;

   regout : process (clk, rst)
   begin
   if (rst = '1') then
      intIOAout <= "00000000";
      intIOBout <= "00000000";
      intIOCout <= "00000000";
      intIODout <= "00000000";
      intIOEout <= "00000000";
      intIOFout <= "00000000";
      intIOGout <= "00000000";
      intIOHout <= "00000000";
      intIOIout <= "00000000";
      intIOJout <= "00000000";
   elsif ((clk'EVENT) and (clk='1')) then
      if (loadIOAout = '1') then
         intIOAout <= DataReg;
      else
         intIOAout <= intIOAout;
      end if;
      --
      if (loadIOBout = '1') then
         intIOBout <= DataReg;
      else
         intIOBout <= intIOBout;
      end if;
      --
      if (loadIOCout = '1') then
         intIOCout <= DataReg;
      else
         intIOCout <= intIOCout;
      end if;
      --
      if (loadIODout = '1') then
         intIODout <= DataReg;
      else
         intIODout <= intIODout;
      end if;
      --
      if (loadIOEout = '1') then
         intIOEout <= DataReg;
      else
         intIOEout <= intIOEout;
      end if;
      --
      if (loadIOFout = '1') then
         intIOFout <= DataReg;
      else
         intIOFout <= intIOFout;
      end if;
      --
      if (loadIOGout = '1') then
         intIOGout <= DataReg;
      else
         intIOGout <= intIOGout;
      end if;
      --
      if (loadIOHout = '1') then
         intIOHout <= DataReg;
      else
         intIOHout <= intIOHout;
      end if;
      --
      if (loadIOIout = '1') then
         intIOIout <= DataReg;
      else
         intIOIout <= intIOIout;
      end if;
      --
      if (loadIOJout = '1') then
         intIOJout <= DataReg;
      else
         intIOJout <= intIOJout;
      end if;
   end if;
   end process regout;

   -- Generate control signals for loading data to output enable registers
   loadIOAen <= LoadStb and DecodeOEn and DecodeIOA;
   loadIOBen <= LoadStb and DecodeOEn and DecodeIOB;
   loadIOCen <= LoadStb and DecodeOEn and DecodeIOC;
   loadIODen <= LoadStb and DecodeOEn and DecodeIOD;
   loadIOEen <= LoadStb and DecodeOEn and DecodeIOE;
   loadIOFen <= LoadStb and DecodeOEn and DecodeIOF;
   loadIOGen <= LoadStb and DecodeOEn and DecodeIOG;
   loadIOHen <= LoadStb and DecodeOEn and DecodeIOH;
   loadIOIen <= LoadStb and DecodeOEn and DecodeIOI;
   loadIOJen <= LoadStb and DecodeOEn and DecodeIOJ;

   regen : process (clk, rst)
   begin
   if (rst = '1') then
      intIOAen <= "00000000";
      intIOBen <= "00000000";
      intIOCen <= "00000000";
      intIODen <= "00000000";
      intIOEen <= "00000000";
      intIOFen <= "00000000";
      intIOGen <= "00000000";
      intIOHen <= "00000000";
      intIOIen <= "00000000";
      intIOJen <= "00000000";
   elsif ((clk'EVENT) and (clk='1')) then
      if (loadIOAen = '1') then
         intIOAen <= DataReg;
      else
         intIOAen <= intIOAen;
      end if;
      --
      if (loadIOBen = '1') then
         intIOBen <= DataReg;
      else
         intIOBen <= intIOBen;
      end if;
      --
      if (loadIOCen = '1') then
         intIOCen <= DataReg;
      else
         intIOCen <= intIOCen;
      end if;
      --
      if (loadIODen = '1') then
         intIODen <= DataReg;
      else
         intIODen <= intIODen;
      end if;
      --
      if (loadIOEen = '1') then
         intIOEen <= DataReg;
      else
         intIOEen <= intIOEen;
      end if;
      if (loadIOFen = '1') then
         intIOFen <= DataReg;
      else
         intIOFen <= intIOFen;
      end if;
      --
      if (loadIOGen = '1') then
         intIOGen <= DataReg;
      else
         intIOGen <= intIOGen;
      end if;
      --
      if (loadIOHen = '1') then
         intIOHen <= DataReg;
      else
         intIOHen <= intIOHen;
      end if;
      --
      if (loadIOIen = '1') then
         intIOIen <= DataReg;
      else
         intIOIen <= intIOIen;
      end if;
      --
      if (loadIOJen = '1') then
         intIOJen <= DataReg;
      else
         intIOJen <= intIOJen;
      end if;
   end if;
   end process regen;

--------------------------------------------------------------------------------
-- Tx State Machine - controls writing of FT232 Data and WR signals
--------------------------------------------------------------------------------
   loadFT232D <= LoadStb and (DecodeRead or not LoadCommand);

   TxStatep : process (clk, rst)
   begin
   if (rst = '1') then
      TxState     <= TxIDLE;
      nWR          <= '0';
      intFT232DEn <= '0';
      senddata    <= '0';
   elsif ((clk'EVENT) and (clk='1')) then
      case TxState is
         -- Wait for Read command
         when TxIDLE =>
            if (loadFT232D = '1') then
               TxState     <= TxWAIT1;
               intFT232DEn <= '1';
            else
               TxState     <= TxIDLE;
               intFT232DEn <= '0';
            end if;
            nWR       <= '0';
            senddata <= '0';
         -- Start of cycle to send Command Data to FT245
         -- Wait for Tx FIFO to signal that it can send data
         when TxWAIT1 =>
            if (intTXE = '1') then
               TxState <= TxCMD1;
               nWR      <= '1';
            else
               TxState <= TxWAIT1;
               nWR      <= '0';
            end if;
            intFT232DEn <= '1';
            senddata    <= '0';
         -- WR is set high for two cycles for the write cycle
         when TxCMD1 =>
            TxState     <= TxCMD2;
            nWR          <= '1';
            intFT232DEn <= '1';
            senddata    <= '0';
         -- Remain in this state until Tx FIFO indicates Wr cycle complete
         -- When this happens return to IDLE if Command is not valid or go to
         -- Data wait state otherwise
         when TxCMD2 =>
            if (intTXE = '1') then
               TxState     <= TxCMD2;
               intFT232DEn <= '1';
               senddata    <= '0';
            elsif (LoadCommand = '0') then
               TxState     <= TxIDLE;
               intFT232DEn <= '0';
               senddata    <= '0';
            else
               TxState     <= TxWAIT2;
               intFT232DEn <= '0';
               senddata    <= '1';
            end if;
            nWR <= '0';
         -- Start of cycle to send Port Data to FT245
         -- Wait for Tx FIFO to signal that it can send data
         when TxWAIT2 =>
            if (intTXE = '1') then
               TxState <= TxDATA1;
               nWR      <= '1';
            else
               TxState <= TxWAIT2;
               nWR      <= '0';
            end if;
            intFT232DEn <= '1';
            senddata    <= '0';
         -- WR is set high for two cycles for the write cycle
         when TxDATA1 =>
            TxState     <= TxDATA2;
            nWR          <= '1';
            intFT232DEn <= '1';
            senddata    <= '0';
         -- Remain in this state until Tx FIFO indicates Wr cycle complete
         when TxDATA2 =>
            if (intTXE = '0') then
               TxState     <= TxIDLE;
               intFT232DEn <= '0';
            else
               TxState     <= TxDATA2;
               intFT232DEn <= '1';
            end if;
            senddata <= '0';
            nWR       <= '0';
         -- default state - return to IDLE
         when others =>
            TxState     <= TxIDLE;
            nWR          <= '0';
            intFT232DEn <= '0';
            senddata    <= '0';
      end case;
   end if;
   end process TxStatep;

--------------------------------------------------------------------------------
-- Register FT245 output data to select either the coded Read command, echo of
-- sync commands or the desired input port
--------------------------------------------------------------------------------
   regFT232Dout : process (clk, rst)
   begin
   if (rst = '1') then
      intFT232Dout <= "00000000";
   elsif ((clk'EVENT) and (clk='1')) then
      -- Otherwise send inverted command with the port bits when doing a read
      if (loadFT232D = '1') then
         if (LoadCommand = '1') then
            intFT232Dout(0) <= CmdReg(0);
            intFT232Dout(1) <= CmdReg(1);
            intFT232Dout(2) <= CmdReg(2);
            intFT232Dout(3) <= CmdReg(3);
            intFT232Dout(4) <= CmdReg(4);
            intFT232Dout(5) <= not CmdReg(5);
            intFT232Dout(6) <= not CmdReg(6);
            intFT232Dout(7) <= not CmdReg(7);
         else
            intFT232Dout <= CmdReg;
         end if;
      -- Send the data on the selected input port when the data word is needed
      elsif (SendData = '1') then
         intFT232Dout(7 downto 0) <= int_Din;
      else
         intFT232Dout <= intFT232Dout;
      end if;
   end if;
   end process regFT232Dout;

   -- Chooses which input port to output to the FT245 chip
   port_sel(0) <= DecodeIOA;
   port_sel(1) <= DecodeIOB;
   port_sel(2) <= DecodeIOC;
   port_sel(3) <= DecodeIOD;
   port_sel(4) <= DecodeIOE;
   port_sel(5) <= DecodeIOF;
   port_sel(6) <= DecodeIOG;
   port_sel(7) <= DecodeIOH;
   port_sel(8) <= DecodeIOI;
   port_sel(9) <= DecodeIOJ;

   mux_Din : process(port_sel,
                     intIOAin, intIOBin, intIOCin, intIODin, intIOEin, 
                     intIOFin, intIOGin, intIOHin, intIOIin, intIOJin) 
   begin
     case port_sel is
       when "0000000001" => int_Din <= intIOAin;
       when "0000000010" => int_Din <= intIOBin;
       when "0000000100" => int_Din <= intIOCin;
       when "0000001000" => int_Din <= intIODin;
       when "0000010000" => int_Din <= intIOEin;
       when "0000100000" => int_Din <= intIOFin;
       when "0001000000" => int_Din <= intIOGin;
       when "0010000000" => int_Din <= intIOHin;
       when "0100000000" => int_Din <= intIOIin;
       when others  => int_Din <= intIOJin;
     end case;
   end process mux_Din;

--------------------------------------------------------------------------------
-- Assign other outputs
--------------------------------------------------------------------------------
   -- The following pins can be configured as clock outputs so use clk_sel
   -- signals to choose what is output to the pins.
   IOA0_muxout : process(clk_selA, cnt, clk, intIOAout(0)) begin
      case clk_selA is
         when "1101" => IOAout(0) <= cnt(11);
         when "1100" => IOAout(0) <= cnt(10);
         when "1011" => IOAout(0) <= cnt(9);
         when "1010" => IOAout(0) <= cnt(8);
         when "1001" => IOAout(0) <= cnt(7);
         when "1000" => IOAout(0) <= cnt(6);
         when "0111" => IOAout(0) <= cnt(5);
         when "0110" => IOAout(0) <= cnt(4);
         when "0101" => IOAout(0) <= cnt(3);
         when "0100" => IOAout(0) <= cnt(2);
         when "0011" => IOAout(0) <= cnt(1);
         when "0010" => IOAout(0) <= cnt(0);
         when "0001" => IOAout(0) <= clk;
         when others => IOAout(0) <= intIOAout(0);
      end case;
   end process IOA0_muxout;

   IOB1_muxout : process(clk_selB, cnt, clk, intIOBout(1)) begin
      case clk_selB is
         when "1101" => IOBout(1) <= cnt(11);
         when "1100" => IOBout(1) <= cnt(10);
         when "1011" => IOBout(1) <= cnt(9);
         when "1010" => IOBout(1) <= cnt(8);
         when "1001" => IOBout(1) <= cnt(7);
         when "1000" => IOBout(1) <= cnt(6);
         when "0111" => IOBout(1) <= cnt(5);
         when "0110" => IOBout(1) <= cnt(4);
         when "0101" => IOBout(1) <= cnt(3);
         when "0100" => IOBout(1) <= cnt(2);
         when "0011" => IOBout(1) <= cnt(1);
         when "0010" => IOBout(1) <= cnt(0);
         when "0001" => IOBout(1) <= clk;
         when others => IOBout(1) <= intIOBout(1);
      end case;
   end process IOB1_muxout;

   IOC0_muxout : process(clk_selC, cnt, clk, intIOCout(0)) begin
      case clk_selC is
         when "1101" => IOCout(2) <= cnt(11);
         when "1100" => IOCout(2) <= cnt(10);
         when "1011" => IOCout(2) <= cnt(9);
         when "1010" => IOCout(2) <= cnt(8);
         when "1001" => IOCout(2) <= cnt(7);
         when "1000" => IOCout(2) <= cnt(6);
         when "0111" => IOCout(2) <= cnt(5);
         when "0110" => IOCout(2) <= cnt(4);
         when "0101" => IOCout(2) <= cnt(3);
         when "0100" => IOCout(2) <= cnt(2);
         when "0011" => IOCout(2) <= cnt(1);
         when "0010" => IOCout(2) <= cnt(0);
         when "0001" => IOCout(2) <= clk;
         when others => IOCout(2) <= intIOCout(2);
      end case;
   end process IOC0_muxout;

   IOD0_muxout : process(clk_selD, cnt, clk, intIODout(0)) begin
      case clk_selD is
         when "1101" => IODout(4) <= cnt(11);
         when "1100" => IODout(4) <= cnt(10);
         when "1011" => IODout(4) <= cnt(9);
         when "1010" => IODout(4) <= cnt(8);
         when "1001" => IODout(4) <= cnt(7);
         when "1000" => IODout(4) <= cnt(6);
         when "0111" => IODout(4) <= cnt(5);
         when "0110" => IODout(4) <= cnt(4);
         when "0101" => IODout(4) <= cnt(3);
         when "0100" => IODout(4) <= cnt(2);
         when "0011" => IODout(4) <= cnt(1);
         when "0010" => IODout(4) <= cnt(0);
         when "0001" => IODout(4) <= clk;
         when others => IODout(4) <= intIODout(4);
      end case;
   end process IOD0_muxout;

   IOE0_muxout : process(clk_selE, cnt, clk, intIOEout(0)) begin
      case clk_selE is
         when "1101" => IOEout(4) <= cnt(11);
         when "1100" => IOEout(4) <= cnt(10);
         when "1011" => IOEout(4) <= cnt(9);
         when "1010" => IOEout(4) <= cnt(8);
         when "1001" => IOEout(4) <= cnt(7);
         when "1000" => IOEout(4) <= cnt(6);
         when "0111" => IOEout(4) <= cnt(5);
         when "0110" => IOEout(4) <= cnt(4);
         when "0101" => IOEout(4) <= cnt(3);
         when "0100" => IOEout(4) <= cnt(2);
         when "0011" => IOEout(4) <= cnt(1);
         when "0010" => IOEout(4) <= cnt(0);
         when "0001" => IOEout(4) <= clk;
         when others => IOEout(4) <= intIOEout(4);
      end case;
   end process IOE0_muxout;

      IOF0_muxout : process(clk_selF, cnt, clk, intIOFout(0)) begin
      case clk_selF is
         when "1101" => IOFout(0) <= cnt(11);
         when "1100" => IOFout(0) <= cnt(10);
         when "1011" => IOFout(0) <= cnt(9);
         when "1010" => IOFout(0) <= cnt(8);
         when "1001" => IOFout(0) <= cnt(7);
         when "1000" => IOFout(0) <= cnt(6);
         when "0111" => IOFout(0) <= cnt(5);
         when "0110" => IOFout(0) <= cnt(4);
         when "0101" => IOFout(0) <= cnt(3);
         when "0100" => IOFout(0) <= cnt(2);
         when "0011" => IOFout(0) <= cnt(1);
         when "0010" => IOFout(0) <= cnt(0);
         when "0001" => IOFout(0) <= clk;
         when others => IOFout(0) <= intIOFout(0);
      end case;
   end process IOF0_muxout;

      IOG0_muxout : process(clk_selG, cnt, clk, intIOGout(0)) begin
      case clk_selG is
         when "1101" => IOGout(0) <= cnt(11);
         when "1100" => IOGout(0) <= cnt(10);
         when "1011" => IOGout(0) <= cnt(9);
         when "1010" => IOGout(0) <= cnt(8);
         when "1001" => IOGout(0) <= cnt(7);
         when "1000" => IOGout(0) <= cnt(6);
         when "0111" => IOGout(0) <= cnt(5);
         when "0110" => IOGout(0) <= cnt(4);
         when "0101" => IOGout(0) <= cnt(3);
         when "0100" => IOGout(0) <= cnt(2);
         when "0011" => IOGout(0) <= cnt(1);
         when "0010" => IOGout(0) <= cnt(0);
         when "0001" => IOGout(0) <= clk;
         when others => IOGout(0) <= intIOGout(0);
      end case;
   end process IOG0_muxout;

      IOH0_muxout : process(clk_selH, cnt, clk, intIOHout(0)) begin
      case clk_selH is
         when "1101" => IOHout(0) <= cnt(11);
         when "1100" => IOHout(0) <= cnt(10);
         when "1011" => IOHout(0) <= cnt(9);
         when "1010" => IOHout(0) <= cnt(8);
         when "1001" => IOHout(0) <= cnt(7);
         when "1000" => IOHout(0) <= cnt(6);
         when "0111" => IOHout(0) <= cnt(5);
         when "0110" => IOHout(0) <= cnt(4);
         when "0101" => IOHout(0) <= cnt(3);
         when "0100" => IOHout(0) <= cnt(2);
         when "0011" => IOHout(0) <= cnt(1);
         when "0010" => IOHout(0) <= cnt(0);
         when "0001" => IOHout(0) <= clk;
         when others => IOHout(0) <= intIOHout(0);
      end case;
   end process IOH0_muxout;


      IOI0_muxout : process(clk_selI, cnt, clk, intIOIout(0)) begin
      case clk_selI is
         when "1101" => IOIout(0) <= cnt(11);
         when "1100" => IOIout(0) <= cnt(10);
         when "1011" => IOIout(0) <= cnt(9);
         when "1010" => IOIout(0) <= cnt(8);
         when "1001" => IOIout(0) <= cnt(7);
         when "1000" => IOIout(0) <= cnt(6);
         when "0111" => IOIout(0) <= cnt(5);
         when "0110" => IOIout(0) <= cnt(4);
         when "0101" => IOIout(0) <= cnt(3);
         when "0100" => IOIout(0) <= cnt(2);
         when "0011" => IOIout(0) <= cnt(1);
         when "0010" => IOIout(0) <= cnt(0);
         when "0001" => IOIout(0) <= clk;
         when others => IOIout(0) <= intIOIout(0);
      end case;
   end process IOI0_muxout;

         IOJ0_muxout : process(clk_selJ, cnt, clk, intIOJout(0)) begin
      case clk_selJ is
         when "1101" => IOJout(0) <= cnt(11);
         when "1100" => IOJout(0) <= cnt(10);
         when "1011" => IOJout(0) <= cnt(9);
         when "1010" => IOJout(0) <= cnt(8);
         when "1001" => IOJout(0) <= cnt(7);
         when "1000" => IOJout(0) <= cnt(6);
         when "0111" => IOJout(0) <= cnt(5);
         when "0110" => IOJout(0) <= cnt(4);
         when "0101" => IOJout(0) <= cnt(3);
         when "0100" => IOJout(0) <= cnt(2);
         when "0011" => IOJout(0) <= cnt(1);
         when "0010" => IOJout(0) <= cnt(0);
         when "0001" => IOJout(0) <= clk;
         when others => IOJout(0) <= intIOJout(0);
      end case;
   end process IOJ0_muxout;

   IOAout(7 downto 1) <= intIOAout(7 downto 1);
   IOBout(7 downto 2) <= intIOBout(7 downto 2);
   IOBout(0) <= intIOBout(0);
   IOCout(7 downto 3) <= intIOCout(7 downto 3);
   IOCout(1 downto 0) <= intIOCout(1 downto 0);
   IODout(7 downto 5) <= intIODout(7 downto 5);
   IODout(3 downto 0) <= intIODout(3 downto 0);
   IOEout(7 downto 5) <= intIOEout(7 downto 5);
   IOEout(3 downto 0) <= intIOEout(3 downto 0);
   IOFout(7 downto 1) <= intIOFout(7 downto 1);
   IOGout(7 downto 1) <= intIOGout(7 downto 1);
   IOHout(7 downto 1) <= intIOHout(7 downto 1);
   IOIout(7 downto 1) <= intIOIout(7 downto 1);
   IOJout(7 downto 1) <= intIOJout(7 downto 1);
   IOAen  <= intIOAen(7 downto 0);
   IOBen  <= intIOBen(7 downto 0);
   IOCen  <= intIOCen(7 downto 0);
   IODen  <= intIODen(7 downto 0);
   IOEen  <= intIOEen(7 downto 0);
   IOFen  <= intIOFen(7 downto 0);
   IOGen  <= intIOGen(7 downto 0);
   IOHen  <= intIOHen(7 downto 0);
   IOIen  <= intIOIen(7 downto 0);
   IOJen  <= intIOJen(7 downto 0);

   FT232DEn  <= intFT232DEn;
   FT232Dout <= intFT232Dout;

   nRD <= not intRD;

END rtl;

